/*
	ProWesS, PROGS Windowing System
	PROforma, PROGS Font & Raster Manager

	PFconfig : utility program for ProWesS package
		allows easy configuration of PROforma

	written by Joachim Van der Auwera for PROGS
	(c) 1996 by PROGS, Professional & Graphical Software
	29.09.1996 started
	19.10.1996 default driver now filled in from CLines if exists
*/

#include <ctype.h>
#include "PFConfig.h"

static double ConvUnit[4]={1.0,25.4/72,1.0/72,1.0};
static char *Unit[4]={"","mm","in","pt"};
static pt PageSizeA4[2]    ={ short2pt(595) , short2pt(842) };
static pt PageSizeA3[2]    ={ short2pt(842) , short2pt(1190)};
static pt PageSizeLetter[2]={ short2pt(612) , short2pt(792) };
static pt PageSizeLegal[2] ={ short2pt(612) , short2pt(1008)};

/* === configure a printer driver === */

static Error PageSizeStrings(Global *g)
{
	char string[MAX_UNITSTRING];

	STRFormatPut(string,"%.2f%s",pt2double(g->xsize)*ConvUnit[g->unit],Unit[g->unit]);
	PWChange(g->oxsize,PW_EDLINE_SET,string,NULL);
	STRFormatPut(string,"%.2f%s",pt2double(g->ysize)*ConvUnit[g->unit],Unit[g->unit]);
	PWChange(g->oysize,PW_EDLINE_SET,string,NULL);
	STRFormatPut(string,"%.2f%s",pt2double(g->xmarg)*ConvUnit[g->unit],Unit[g->unit]);
	PWChange(g->oxmarg,PW_EDLINE_SET,string,NULL);
	STRFormatPut(string,"%.2f%s",pt2double(g->ymarg)*ConvUnit[g->unit],Unit[g->unit]);
	PWChange(g->oymarg,PW_EDLINE_SET,string,NULL);
	return ERR_OK;
}

static Error UnitSet(PWObject object)
{
	PWObject *units;
	Global *g;

	PWQuery(object,PW_GLOBAL_AUXILIARY,&g);
	PWQuery(object,PW_OBJECT_AUXILIARY,&units);
	if (object==units[0]) g->unit=UNIT_MM;
	if (object==units[1]) g->unit=UNIT_IN;
	if (object==units[2]) g->unit=UNIT_PT;
	PWChange(units[0],PW_LOOSE_STATUS,
				g->unit==UNIT_MM?PW_STATUS_SELECTED:PW_STATUS_AVAILABLE,NULL);
	PWChange(units[1],PW_LOOSE_STATUS,
				g->unit==UNIT_IN?PW_STATUS_SELECTED:PW_STATUS_AVAILABLE,NULL);
	PWChange(units[2],PW_LOOSE_STATUS,
				g->unit==UNIT_PT?PW_STATUS_SELECTED:PW_STATUS_AVAILABLE,NULL);
	return PageSizeStrings(g);
}

static Error UnitTest(PWObject object, EdlineInfo *info)
{
	char key=info->key;
	int len;

	if (key=='+' || key=='-')
	{
		if (info->string[0]=='+' || info->string[0]=='-')
		{
			info->string[0]=key;
			return PW_EDLINE_TEST_CHANGE;
		}
		if (info->length+1<info->maxlength)
		{
			MEMCopy(info->maxlength-1,info->string,info->string+1);
			info->string[0]=key;
			info->cursor++;
			return PW_EDLINE_TEST_CHANGE;
		}
		return PW_EDLINE_TEST_NOACCEPT;
	}
	if (key=='.')
	{
		/* only accept if no decimal point yet */
		char *str;
		for (str=info->string ; *str ; str++)
			if (*str=='.') return PW_EDLINE_TEST_NOACCEPT;
		return PW_EDLINE_TEST_ACCEPT;
	}
	len=info->length-2;
	switch(key)
	{
	case 'p':
		info->string[len  ]='p';
		info->string[len+1]='t';
		return PW_EDLINE_TEST_CHANGE;
	case 'm':
		info->string[len  ]='m';
		info->string[len+1]='m';
		return PW_EDLINE_TEST_CHANGE;
	case 'i':
		info->string[len  ]='i';
		info->string[len+1]='n';
		return PW_EDLINE_TEST_CHANGE;
	case 'c':
		info->string[len  ]='c';
		info->string[len+1]='m';
		return PW_EDLINE_TEST_CHANGE;
	case 0:
		info->cursor-=2;
		if (info->cursor<0) info->cursor=0;
		return PW_EDLINE_TEST_CHANGE;
	case 200: /* right cursor key */
		return (info->cursor>=len) ? PW_EDLINE_TEST_NOACCEPT :
									 PW_EDLINE_TEST_ACCEPT;
	default:
		if (isprint(key) && !isdigit(key))
			return PW_EDLINE_TEST_NOACCEPT;
		else
			return PW_EDLINE_TEST_ACCEPT;
	}
}

static Error UnitFill(PWObject object)
{
	char value[MAX_UNITSTRING];
	pt *result; double db; int count;

	PWChange(object,PW_EDLINE_GET,MAX_UNITSTRING,value,NULL);
	PWQuery(object,PW_OBJECT_AUXILIARY,&result);
	STRFormatGet(value,NULL,"%f%n",&db,&count);
	switch(value[count])
	{
	case 'm':
		db*=72/25.4;
		break;
	case 'i':
		db*=72;
		break;
	case 'c':
		db*=72/2.54;
		break;
	}
	*result=double2pt(db);
	return ERR_OK;
}

static EdlineSet(Global *g, PWObject win, char *label, char key,
				 PWObject *new, pt *value)
{
	Error err;
	char string[MAX_UNITSTRING];

	STRFormatPut(string,"%.2f%s",pt2double(*value)*ConvUnit[g->unit],Unit[g->unit]);
	catch( PWCreate(win,NULL,PW_TYPE_LABEL,
					PW_LABEL_TEXT, label,
					PW_SCALE_FACTOR, 0,
					PW_POSITION_NEXT_ROW,
					NULL) );
	catch( PWCreate(win,new,PW_TYPE_EDLINE,
					PW_EDLINE_MAXLENGTH, MAX_UNITSTRING,
					PW_EDLINE_SET, string,
					PW_EDLINE_KEYPRESS, key,
					PW_EDLINE_ACTION_TEST, UnitTest,
					PW_EDLINE_ACTION_AFTER, UnitFill,
					PW_OBJECT_AUXILIARY, value,
					NULL) );
	return err;
}

static Error PageSizeSet(PWObject object)
{
	Global *g;
	pt *size, *value;
	char string[MAX_UNITSTRING];

	PWQuery(object,PW_GLOBAL_AUXILIARY,&g);
	PWQuery(object,PW_OBJECT_AUXILIARY,&size);

	g->xsize=size[0];
	g->ysize=size[1];
	g->xmarg=g->ymarg=0;
	return PageSizeStrings(g);
}

static Error Get(PWObject object)
{
	Global *g;
	Error err;
	Gstate gstate;
	Bbox page={short2pt(2000),short2pt(2000),0,0};

	PWQuery(object,PW_GLOBAL_AUXILIARY,&g);

	err=PFGstateInit(PF_NAME(g->driver),"NULL",&page,&gstate,NULL);
	if (err) return DisplayError(err);
	PFPageBboxFullGet(gstate,&page);
	PFGstateRemove(gstate);
	g->xsize=page.xsiz;
	g->ysize=page.ysiz;
	g->xmarg=page.xorg;
	g->ymarg=page.yorg;
	return PageSizeStrings(g);
}

static Error Set(PWObject object)
{
	Global *g;
	Error err;
	char string[MAX_LINE];

	PWQuery(object,PW_GLOBAL_AUXILIARY,&g);
	PWChange(g->odevice,PW_EDLINE_GET,IO_MAXFULLNAME,g->device,NULL);

	STRFormatPutSafe(string,MAX_LINE,"d %s",g->driver);
	err=THINGCall(PF_THING_NAME,PF_THING_CNFG,1,string);
	if (err) return DisplayError(err);
	STRFormatPutSafe(string,MAX_LINE,"DEFAULT-DEVICE %s",g->device);
	err=THINGCall(PF_THING_NAME,PF_THING_CNFG,1,string);
	if (err) return DisplayError(err);
	STRFormatPutSafe(string,MAX_LINE,"PRINTABLE-AREA-SIZE %f %f",
				pt2double(g->xsize),pt2double(g->ysize));
	err=THINGCall(PF_THING_NAME,PF_THING_CNFG,1,string);
	if (err) return DisplayError(err);
	STRFormatPutSafe(string,MAX_LINE,"PRINTABLE-AREA-ORIGIN %f %f",
				pt2double(g->xmarg),pt2double(g->ymarg));
	err=THINGCall(PF_THING_NAME,PF_THING_CNFG,1,string);
	if (err) return DisplayError(err);

	/* now also store this in the internal structures */
	if (!g->current)
	{
		/* allocate a new CLines for this driver */
		catch( MEMUserAllocate(g->lines,&g->current) );
		STRCopy(g->driver,g->current->driver);
		g->current->next=g->config;
		g->config=g->current;
	}
	STRCopy(g->device,g->current->device);
	g->current->xsize=g->xsize;
	g->current->ysize=g->ysize;
	g->current->xmarg=g->xmarg;
	g->current->ymarg=g->ymarg;
	g->changed=TRUE;

	return ERR_OK;
}

static Error Try(PWObject object)
{
	Global *g;
	Error err;
	Gstate gstate;
	Bbox page={short2pt(2000),short2pt(2000),0,0};
	int passes;

	PWQuery(object,PW_GLOBAL_AUXILIARY,&g);

	err=PFGstateInit(PF_NAME(g->driver),NULL,&page,&gstate,&passes);
	if (err) return DisplayError(err);
	PFPageBboxFullGet(gstate,&page);

	/* draw two boxes in top left and bottom right corner */
	/* and two lines, each 4 inches from page border */
	for ( ; passes ; passes--)
	{
		/* draw two boxes in top left and bottom right corner */
		/* and two lines, each 4 inches from page border */
		pt dist=max(short2pt(288),4*page.xorg);
		PFColourGray(gstate,short2pt(100));
		PFFontScale(gstate,short2pt(10));
		PFPathStroked(gstate);
		PFLineWidth(gstate,pt_half);
		PFMoveTo(gstate,short2pt(72*4),0);
		PFLineR(gstate,0,page.ysiz+dist);
		PFPathDraw(gstate);
		PFMoveTo(gstate,0,short2pt(72*4));
		PFLineR(gstate,page.xsiz+dist,0);
		PFPathDraw(gstate);
		PFMoveTo(gstate,short2pt(72*4+1),short2pt(72*4+10));
		PFStringShow(gstate,"lines at 4 inches from edge of page");
		PFMoveTo(gstate,double2pt(340.157),0);
		PFLineR(gstate,0,page.ysiz+dist);
		PFPathDraw(gstate);
		PFMoveTo(gstate,0,double2pt(340.157));
		PFLineR(gstate,page.xsiz+dist,0);
		PFPathDraw(gstate);
		PFMoveTo(gstate,double2pt(340.157+1),double2pt(340.157+10));
		PFStringShow(gstate,"lines at 12cm from edge of page");

		PFPageShow(gstate);
	}
	return PFGstateRemove(gstate);
}

Error PrinterConfig(Global *g)
{
	Error err;
	PWObject win, units[3];
	CLines *curr=g->config;

	catch( DriverSelect(NULL,FALSE,FALSE,g->driver) );

	/* try to find a CLines for this driver */
	while (curr && !STRSameCD(curr->driver,g->driver)) curr=curr->next;
	g->current=curr;
	if (g->current) STRCopy(g->current->device,g->device);

	catch( PWCreate(NULL,&win,PW_TYPE_OUTLINE,
					PW_OUTLINE_TITLE_TEXT, "dconfig - driver config",
					PW_OUTLINE_QUIT,
					PW_OUTLINE_QUIT_KEYPRESS, 27,
					PW_OUTLINE_SLEEP,
					PW_OUTLINE_HELP,
					PW_OUTLINE_HELP_DIRECTORY, "$PWSDIR_doc_",
					PW_OUTLINE_HELP_FILE, "PFconfig_html",
					PW_OUTLINE_HELP_POSITION, "cdriv",
					PW_GLOBAL_AUXILIARY, g,
					PW_WINDOW_NOSCALE,
					NULL) );

	catch( PWCreate(win,NULL,PW_TYPE_INFOSTRING,
					PW_INFOSTRING_TEXT, g->driver,
					NULL) );

	catch( PWCreate(win,NULL,PW_TYPE_LABEL,
					PW_LABEL_TEXT, "unit",
					PW_SCALE_FACTOR, 0,
					PW_POSITION_NEXT_ROW,
					NULL) );
	catch( PWCreate(win,&units[0],PW_TYPE_LOOSE_ITEM,
					PW_LOOSE_TEXT, "mm",
					PW_LOOSE_KEYPRESS, 'm',
					PW_LOOSE_ACTION_HIT, UnitSet,
					PW_OBJECT_AUXILIARY, units,
					NULL) );
	catch( PWCreate(win,&units[1],PW_TYPE_LOOSE_ITEM,
					PW_LOOSE_TEXT, "inch",
					PW_LOOSE_KEYPRESS, 'i',
					PW_LOOSE_ACTION_HIT, UnitSet,
					PW_OBJECT_AUXILIARY, units,
					NULL) );
	catch( PWCreate(win,&units[2],PW_TYPE_LOOSE_ITEM,
					PW_LOOSE_TEXT, "pt",
					PW_LOOSE_KEYPRESS, 'p',
					PW_LOOSE_ACTION_HIT, UnitSet,
					PW_OBJECT_AUXILIARY, units,
					NULL) );

	catch( PWCreate(win,NULL,PW_TYPE_LABEL,
					PW_LABEL_TEXT, "default device",
					PW_SCALE_FACTOR, 0,
					PW_POSITION_NEXT_ROW,
					NULL) );
	catch( PWCreate(win,&g->odevice,PW_TYPE_EDLINE,
					PW_EDLINE_MAXLENGTH, MAX_UNITSTRING,
					PW_EDLINE_SET, g->device,
					PW_EDLINE_KEYPRESS, 'd',
					NULL) );

	catch( EdlineSet(g,win,"width of page",'w',&g->oxsize,&g->xsize) );
	catch( EdlineSet(g,win,"height of page",'h',&g->oysize,&g->ysize) );
	catch( EdlineSet(g,win,"left margin",'l',&g->oxmarg,&g->xmarg) );
	catch( EdlineSet(g,win,"top margin",'t',&g->oymarg,&g->ymarg) );

	catch( PWCreate(win,NULL,PW_TYPE_LABEL,
					PW_LABEL_TEXT, "reset page size",
					PW_SCALE_FACTOR, 0,
					PW_POSITION_NEXT_ROW,
					NULL) );
	catch( PWCreate(win,NULL,PW_TYPE_LOOSE_ITEM,
					PW_LOOSE_TEXT, "A4",
					PW_LOOSE_KEYPRESS, '4',
					PW_LOOSE_ACTION_HIT, PageSizeSet,
					PW_LOOSE_CHANGE_STATUS, FALSE,
					PW_OBJECT_AUXILIARY, &PageSizeA4,
					NULL) );
	catch( PWCreate(win,NULL,PW_TYPE_LOOSE_ITEM,
					PW_LOOSE_TEXT, "A3",
					PW_LOOSE_KEYPRESS, '3',
					PW_LOOSE_ACTION_HIT, PageSizeSet,
					PW_LOOSE_CHANGE_STATUS, FALSE,
					PW_OBJECT_AUXILIARY, &PageSizeA3,
					NULL) );
	catch( PWCreate(win,NULL,PW_TYPE_LOOSE_ITEM,
					PW_LOOSE_TEXT, "letter",
					PW_LOOSE_ACTION_HIT, PageSizeSet,
					PW_LOOSE_CHANGE_STATUS, FALSE,
					PW_OBJECT_AUXILIARY, &PageSizeLetter,
					NULL) );
	catch( PWCreate(win,NULL,PW_TYPE_LOOSE_ITEM,
					PW_LOOSE_TEXT, "legal",
					PW_LOOSE_ACTION_HIT, PageSizeSet,
					PW_LOOSE_CHANGE_STATUS, FALSE,
					PW_OBJECT_AUXILIARY, &PageSizeLegal,
					NULL) );

	catch( PWCreate(win,NULL,PW_TYPE_LOOSE_ITEM,
					PW_LOOSE_CHANGE_STATUS, FALSE,
					PW_LOOSE_TEXT, "Get",
					PW_LOOSE_KEYPRESS, 'g',
					PW_LOOSE_ACTION_HIT, Get,
					PW_POSITION_NEXT_ROW,
					NULL) );
	catch( PWCreate(win,NULL,PW_TYPE_LOOSE_ITEM,
					PW_LOOSE_CHANGE_STATUS, FALSE,
					PW_LOOSE_TEXT, "Set",
					PW_LOOSE_KEYPRESS, 's',
					PW_LOOSE_ACTION_HIT, Set,
					NULL) );
	catch( PWCreate(win,NULL,PW_TYPE_LOOSE_ITEM,
					PW_LOOSE_CHANGE_STATUS, FALSE,
					PW_LOOSE_TEXT, "Try",
					PW_LOOSE_KEYPRESS, 't',
					PW_LOOSE_ACTION_HIT, Try,
					NULL) );

	UnitSet(units[g->unit-1]);
	catch( Get(win) );

	err=PWActivate(win);
	PWRemove(win);
	return err;
}
