/*
	ProWesS, PROGS Windowing System
	PROforma, PROGS Font & Raster Manager

	PFconfig : utility program for ProWesS package
		allows easy configuration of PROforma

	written by Joachim Van der Auwera for PROGS
	(c) 1996 by PROGS, Professional & Graphical Software
	02.10.1996 ConfigRead written
	03.10.1996 ConfigWrite written
*/

#include <ctype.h>
#include "PFConfig.h"

/* === allow permanent changes === */
/* read and write PROforma config file from $PWSDIR_mine_PROforma_cfg */

Error LineAdd(Global *g, char *line, Lines **base)
{
	Error err=ERR_OK;
	Lines *lines=*base;

	if (base==&g->fonts)
		while (lines && !STRSameCD(lines->line,line)) lines=lines->next;
	else
		while (lines && !STRSameCI(lines->line,line)) lines=lines->next;
	if (!lines)
	{
		catch( MEMUserAllocate(g->lines,&lines) );
		lines->next=*base;
		*base=lines;
		MEMCopy(MAX_LINE,line,lines->line);
		g->changed=TRUE;
	}
	return err;
}

static Error ReadDefine(CLines *curr, char *name, char *value)
{
	double x,y;
	if (STRSameCD(name,"DEFAULT-DEVICE"))
	{
		STRCopySafe(value,curr->device,IO_MAXFULLNAME);
		return ERR_OK;
	}
	if (STRSameCD(name,"PRINTABLE-AREA-SIZE"))
	{
		STRFormatGet(value,NULL,"%f %f",&x,&y);
		curr->xsize=double2pt(x);
		curr->ysize=double2pt(y);
		return ERR_OK;
	}
	if (STRSameCD(name,"PRINTABLE-AREA-ORIGIN"))
	{
		STRFormatGet(value,NULL,"%f %f",&x,&y);
		curr->xmarg=double2pt(x);
		curr->ymarg=double2pt(y);
		return ERR_OK;
	}
	return ERR_OK;
}

static Error Read(Global *g, char *line)
{
	Error err=ERR_OK;
	char *pos;
	Lines *base;

	/* just skip empty and comment lines */
	if (line[0]=='\0' || line[0]==';' || line[0]=='%') return err;

	if (line[1]==' ')
	{
		/* command for configurator */
		g->newdef=line[0]=='d'?TRUE:FALSE;
		switch(line[0])
		{
		case 'S':
			/* search path for fonts */
			line+=2;
			if (*line=='+') line++;
			while (*line)
			{
				catch( MEMUserAllocate(g->lines,&base) );
				base->next=g->fpath;
				g->fpath=base;
				for (pos=base->line ; *line && *line!=';' ; ) *pos++=*line++;
				if (*line==';') line++;
				*pos='\0';
			}
			break;
		case 's':
			/* default search path for drivers */
			line+=2;
			if (*line=='+') line++;
			while (*line)
			{
				catch( MEMUserAllocate(g->lines,&base) );
				base->next=g->dpath;
				g->dpath=base;
				for (pos=base->line ; *line && *line!=';' ; ) *pos++=*line++;
				if (*line==';') line++;
				*pos='\0';
			}
			break;
		case 'D':
			/* load a driver */
			catch( MEMUserAllocate(g->lines,&base) );
			base->next=g->drivers;
			g->drivers=base;
			STRCopy(line+2,base->line);
			break;
		case 'C':
			/* set size of fontcache */
			g->cachecount=g->cachesize=0;
			STRFormatGet(line+2,NULL,"%d %d",&g->cachesize,&g->cachecount);
			break;
		case 'c':
			/* set size of colour cache */
			STRFormatGet(line+2,NULL,"%d",&g->colcache);
			if (g->colcache>256) g->colcache=256;
			if (g->colcache<2) g->colcache=1;
			break;
		case 'M':
			/* set memory configuration */
			STRFormatGet(line+2,NULL,"%d",&g->memory);
			break;
		case 'P':
			catch( MEMUserAllocate(g->lines,&base) );
			MEMClear(sizeof(Lines),base);
			base->next=g->fonts;
			g->fonts=base;
			STRCopy(line+2,base->line);
			if (MEMSearch(';',MAX_LINE,base->line,&line)==ERR_OK) *line='\0';
			break;
		case 'R':
			/* add a resident font */
			catch( MEMUserAllocate(g->lines,&base) );
			base->next=g->rfont;
			g->rfont=base;
			STRCopy(line+2,base->line);
			break;
		case 'd':
			/* select the default printer driver */
			STRCopy(line+2,g->printer);
			break;
		default:
			err=ERR_IPAR;
		}
	}
	else
	{
		/* define - PFConfig passes this to last defined default driver */

		if (g->printer[0])
		{
			char *first, *second;
			CLines *base;

			if (g->newdef)
			{
				catch( MEMUserAllocate(g->lines,&base) );
				MEMClear(sizeof(CLines),base);
				base->next=g->config;
				g->config=base;
				STRCopy(g->printer,base->driver);
				g->current=base;
				g->newdef=FALSE;
			}

			/* now this line has to be split in keyword/data */
			second=first=line;
			while (*second && !isspace(*second)) second++;
			if (isspace(*second))
				*second++='\0';
			while (isspace(*second)) second++;

			err=ReadDefine(g->current,first,second);
		}
	}
	return err;
}

Error ConfigRead(Global *g)
{
	Error err;
	Channel chid;
	char line[MAX_LINE];

	catch( IOOpen("$PWSDIR_mine_PROforma_cfg",OPEN_READ,&chid) );
	while (!IOTest(chid))
	{
		IOGetString(chid, line, MAX_LINE);
		catch( Read(g,line) );
	}
	return IOClose(chid);
}

/* routines to save the PROforma_cfg file */

static Error WriteLines(Channel chid, Lines *lines, Error (*wline)(Channel chid, Lines *line))
{
	Error err;
	Lines *prev=NULL, *curr;

	while (prev!=lines)
	{
		for (curr=lines ; curr->next!=prev ; curr=curr->next);
		prev=curr;
		catch( wline(chid,curr) );
	}
	return ERR_OK;
}

static Error ListDrivPath(Channel chid, Lines *line)
{
	char str[MAX_LINE];
	STRFormatPut(str,"s +%s",line->line);
	return IOPutString(chid,str,NULL);
}

static Error ListFontPath(Channel chid, Lines *line)
{
	char str[MAX_LINE];
	STRFormatPut(str,"S +%s",line->line);
	return IOPutString(chid,str,NULL);
}

static Error ListDriv(Channel chid, Lines *line)
{
	char str[MAX_LINE];
	STRFormatPut(str,"D %s",line->line);
	return IOPutString(chid,str,NULL);
}

static Error ListConfig(Channel chid, Lines *_line)
{
	CLines *line=(CLines *)_line;
	char str[MAX_LINE];
	Error err;

	STRFormatPut(str,"d %s",line->driver);
	catch( IOPutString(chid,str,NULL) );
	STRFormatPut(str,"DEFAULT-DEVICE %s",line->device);
	catch( IOPutString(chid,str,NULL) );
	STRFormatPut(str,"PRINTABLE-AREA-SIZE %f %f",
					pt2double(line->xsize),pt2double(line->ysize));
	catch( IOPutString(chid,str,NULL) );
	STRFormatPut(str,"PRINTABLE-AREA-ORIGIN %f %f",
					pt2double(line->xmarg),pt2double(line->ymarg));
	catch( IOPutString(chid,str,NULL) );
	return err;
}

static Error ListFontRes(Channel chid, Lines *line)
{
	char str[MAX_LINE];
	STRFormatPut(str,"R %s",line->line);
	return IOPutString(chid,str,NULL);
}

static Error ListFont(Channel chid, Lines *line)
{
	char str[MAX_LINE];
	STRFormatPut(str,"P %s;%s",line->line,line->line+strlen(line->line)+1);
	return IOPutString(chid,str,NULL);
}

static Error wfile(Channel chid, Global *g)
{
	Error err;
	char line[MAX_LINE];

	catch( IOPutString(chid,"; PROforma, PROGS Font & Raster Manager, configuration file",NULL) );
	catch( IOPutString(chid,"; PROforma, (c) PROGS, Professional & Graphical Software",NULL) );
	catch( IOPutString(chid,"; written by Joachim Van der Auwera",NULL) );
	catch( IOPutString(chid,"; config file written by PFconfig",NULL) );

	catch( IOPutString(chid,"\n; memory options",NULL) );
	STRFormatPut(line,"C %d %d",g->cachesize,g->cachecount);
	catch( IOPutString(chid,line,NULL) );
	STRFormatPut(line,"c %d",g->colcache);
	catch( IOPutString(chid,line,NULL) );
	STRFormatPut(line,"M %d",g->memory);
	catch( IOPutString(chid,line,NULL) );

	catch( IOPutString(chid,"\n; searchpath for drivers",NULL) );
	catch( WriteLines(chid,g->dpath,ListDrivPath) );

	catch( IOPutString(chid,"\n; searchpath for fonts",NULL) );
	catch( WriteLines(chid,g->fpath,ListFontPath) );

	catch( IOPutString(chid,"\n; load drivers (mixed screen, printer, bitmap & picture)",NULL) );
	catch( WriteLines(chid,g->drivers,ListDriv) );
	catch( IOPutString(chid,"\n; configure drivers",NULL) );
	catch( WriteLines(chid,(Lines *)g->config,ListConfig) );
	catch( IOPutString(chid,"\n; set default driver",NULL) );
	STRFormatPut(line,"d %s",g->printer);
	catch( IOPutString(chid,line,NULL) );

	catch( IOPutString(chid,"\n; load resident fonts",NULL) );
	catch( WriteLines(chid,g->rfont,ListFontRes) );

	catch( IOPutString(chid,"\n; list of fonts which can be used",NULL) );
	catch( WriteLines(chid,g->fonts,ListFont) );
	return err;
}

Error ConfigWrite(Global *g)
{
	Error err=ERR_OK;
	Channel chid;

	if (!g->changed) return err;
	catch( IOOpen("$PWSDIR_mine_PROforma_cfg",OPEN_OVER,&chid) );
	err=wfile(chid,g);
	IOClose(chid);
	g->changed=FALSE;
	return err;
}
